signature IL1 =
  sig
    exception Error of string

    (* labels *)

    (* NB: the various "label" constructors correspond roughly to the
       overbar function in HS *)
    datatype label =
        LabTycon of string
      | LabTyvar of string
      | LabExp of string
      | LabModule of string
      | LabFunctor of string
      | LabRecord of string
      | LabSum of string
      | LabInternal of string (* for public, private *)
      | LabStarred of label (* starred label *)

    val string_of_label : label -> string
    val label_eq : label * label -> bool

    type 'a var = 'a Variable.var

    type expvar = Variables.exp var
    type convar = Variables.con var
    type modvar = Variables.mod var

    structure ConVarSet : ORD_SET where type Key.ord_key = convar
    structure ModVarSet : ORD_SET where type Key.ord_key = modvar

    (* abstract syntax *)

    datatype kind =
        KndType
      | KndPi of convar * kind * kind (* convar bound in kind2 *)
      | KndProd of kind list
      | KndSing of con

    and con =
        ConVar of convar
      | ConEvar of ebind ref (* existential variable, for unification *)
      | ConAbs of convar * kind * con (* convar bound in con *)
      | ConApp of con * con
      | ConProd of con list
      | ConProj of int * con
      | ConMu of convar * kind * con (* convar bound in con, kind = KndType^n *)
      | ConModProj of module * label * label list
            (* invariant: module is a value *)
            (* the label list represents the labels in the module *)

      (* cons with kind KndType *)
      | TyLit of Literal.lit_ty
      | TyArrow of con * con
      | TyRef of con
      | TyTagged
      | TyTag of con
      | TySum of (label * con) list
      | TyRecord of (label * con) list
      | TyForall of convar * kind * con (* convar bound in con *)

    (* Evar binding *)
    and ebind =
        EvFree of ConVarSet.set * ModVarSet.set * int
               (* con deps,       module deps,    id *)
               (* invariant: these variables are never
                  substituted into -- fatal error if they are *)
      | EvBound of con (* unified with con *)

    and exp =
        ExVar of expvar
      | ExLit of Literal.lit
      | ExUnOp of Literal.unop * exp
      | ExBinOp of exp * Literal.binop * exp
      | ExFix of (expvar * expvar * con * con * exp) list
      | ExApp of exp * exp
      | ExHandle of exp * exp
      | ExRaise of con * exp
      | ExRef of exp
      | ExGet of exp
      | ExSet of exp * exp
      | ExRoll of con * exp
      | ExUnroll of exp
      | ExRecord of (label * exp) list
      | ExRecordProj of label * exp * label list
      | ExSumInj of con * label * exp * label list
      | ExCase of con * exp * (label * exp) list (* con is result type *)
      | ExTagInj of exp * exp
      | ExNewTag of con
      | ExIfTagOf of exp * exp * exp * exp
      | ExConAbs of convar * kind * exp (* convar bound in exp *)
      | ExConApp of exp * con
      | ExLet of expvar * exp * exp (* expvar bound in exp2 *)
      | ExModLet of modvar * module * exp * con (* modvar bound in exp *)
      | ExModProj of module * label * label list
            (* the label list represents the labels in the module *)

    and module =
        ModVar of modvar
      | ModStruct of (label * bnd) list
      | ModAbs of modvar * signat * module (* modvar bound in module *)
      | ModApp of module * module
      | ModModProj of module * label * label list
      | ModConstrain of module * signat
      | ModLet of modvar * module * module * signat
               (* modvar bound in module2 *)

    and bnd =
        BndExp of expvar * exp
      | BndCon of convar * con
      | BndMod of modvar * module
    (* NB: in a bnd list bnd::bnds, bnd's var bound in bnds *)

    and signat =
        SigStruct of (label * dec) list
      | SigPi of modvar * signat * signat (* modvar bound in signat2 *)

    and dec =
        DecExpCon of expvar * con
      | DecConKind of convar * kind
      | DecModSig of modvar * signat
    (* NB: in a dec list dec::decs, dec's var bound in decs *)

    type fbnd = expvar * expvar * con * con * exp
                   (* f      (x     : ty) : ty  = e *)
                   (* expvar2 (x) bound in exp (e),
                      expvar1 (f) bound in all exps *)

    type sdec = label * dec
    type sbnd = label * bnd

    val labdom : (label * 'a) list -> label list

    (* derived forms *)
    val tyProd : con list -> con
    val exProd : exp list -> exp
    val exProj : int -> exp -> int -> exp

    val tyUnit : con
    val exUnit : exp

    val kndUnit : kind
    val conUnit : con

    val tyInt : con
    val tyChar : con
    val tyString : con

    val exInt : int -> exp
    val exChar : char -> exp
    val exString : string -> exp

    val tyBool : con
    val exFalse : exp
    val exTrue : exp
    val exBool : bool -> exp
    (* con is the result type *)
    val exIf : con -> exp * exp * exp -> exp

    val exLam : expvar * con * con * exp -> exp
    val exPproj : label * con * exp * exp -> exp
    val exLet : bnd list * exp -> exp

    val kndMultiPi : convar list * kind -> kind
    val conMultiAbs : convar list * con -> con

    val exMultiConAbs : convar list * exp -> exp
    val tyMultiForall : convar list * con -> con

    val kndArrow : kind * kind -> kind
    val kndExp : kind * int -> kind

    val sigArrow : signat * signat -> signat

    val conMultiMu : convar list * con -> con

    val labelledKndSing : con -> kind -> kind


    (* alpha variation *)

    (* [alpha_vary (var, phrase) replacefun] yields (var', phrase'), where
       var' is fresh, and (var. phrase) is alpha-equal to (var'. phrase') *)
    val alpha_vary : 'a var * 'b
                   -> ('a var -> 'a var -> 'b -> 'c) -> 'a var * 'c
    val alpha_vary2 : 'a var * 'b -> 'a var * 'b
                   -> ('a var -> 'a var -> 'b -> 'c) -> 'a var * ('c * 'c)
    val alpha_vary_many : 'a var list * 'b list
                       -> ('a var -> 'a var -> 'b -> 'b)
                       -> 'a var list * 'b list

    val alpha_vary_dec_sdecs : dec * sdec list -> dec * sdec list
    val alpha_vary_bnd_sbnds : bnd * sbnd list -> bnd * sbnd list


    (* various replacement functions suitable for use with the above *)

    (* by invariant, if called directly, these should only be called with
       a fresh second convar -- these do not perform alpha-conversion! *)
    val replace_con_con : convar -> convar -> con -> con
    val replace_con_exp : convar -> convar -> exp -> exp
    val replace_con_kind : convar -> convar -> kind -> kind
    val replace_con_module : convar -> convar -> module -> module
    val replace_con_sbnds : convar -> convar
                         -> (label * bnd) list -> (label * bnd) list
    val replace_con_sdecs : convar -> convar
                         -> (label * dec) list -> (label * dec) list
    val replace_con_signat : convar -> convar -> signat -> signat
    val replace_exp_con : expvar -> expvar -> con -> con
    val replace_exp_exp : expvar -> expvar -> exp -> exp
    val replace_exp_kind : expvar -> expvar -> kind -> kind
    val replace_exp_module : expvar -> expvar -> module -> module
    val replace_exp_sbnds : expvar -> expvar
                         -> (label * bnd) list -> (label * bnd) list
    val replace_exp_sdecs : expvar -> expvar
                         -> (label * dec) list -> (label * dec) list
    val replace_exp_signat : expvar -> expvar -> signat -> signat
    val replace_mod_con : modvar -> modvar -> con -> con
    val replace_mod_exp : modvar -> modvar -> exp -> exp
    val replace_mod_kind : modvar -> modvar -> kind -> kind
    val replace_mod_module : modvar -> modvar -> module -> module
    val replace_mod_sbnds : modvar -> modvar
                         -> (label * bnd) list -> (label * bnd) list
    val replace_mod_sdecs : modvar -> modvar
                         -> (label * dec) list -> (label * dec) list
    val replace_mod_signat : modvar -> modvar -> signat -> signat


    (* various substitution functions *)

    (* subst_xxx_phrase substitutes an xxx for an xxxvar in a phrase *)
    val subst_con_con : con -> convar -> con -> con
    val subst_con_exp : con -> convar -> exp -> exp
    val subst_con_kind : con -> convar -> kind -> kind
    val subst_con_module : con -> convar -> module -> module
    val subst_con_sbnds : con -> convar
                       -> (label * bnd) list -> (label * bnd) list
    val subst_con_sdecs : con -> convar
                       -> (label * dec) list -> (label * dec) list
    val subst_con_signat : con -> convar -> signat -> signat

    val subst_exp_con : exp -> expvar -> con -> con
    val subst_exp_exp : exp -> expvar -> exp -> exp
    val subst_exp_kind : exp -> expvar -> kind -> kind
    val subst_exp_module : exp -> expvar -> module -> module
    val subst_exp_sbnds : exp -> expvar
                       -> (label * bnd) list -> (label * bnd) list
    val subst_exp_sdecs : exp -> expvar
                       -> (label * dec) list -> (label * dec) list
    val subst_exp_signat : exp -> expvar -> signat -> signat

    val subst_mod_con : module -> ModVarSet.item -> con -> con
    val subst_mod_exp : module -> modvar -> exp -> exp
    val subst_mod_kind : module -> ModVarSet.item -> kind -> kind
    val subst_mod_module : module -> ModVarSet.item -> module -> module
    val subst_mod_sbnds : module -> modvar
                       -> (label * bnd) list -> (label * bnd) list
    val subst_mod_sdecs : module -> modvar
                       -> (label * dec) list -> (label * dec) list
    val subst_mod_signat : module -> modvar -> signat -> signat

    (* free modvars *)
    val fmv_bnd : bnd -> ModVarSet.set
    val fmv_con : con -> ModVarSet.set
    val fmv_dec : dec -> ModVarSet.set
    val fmv_exp : exp -> ModVarSet.set
    val fmv_kind : kind -> ModVarSet.set
    val fmv_mod : module -> ModVarSet.set
    val fmv_sig : signat -> ModVarSet.set

    (* various mapping functions for manipulating IL terms *)
    val map_con : (convar -> con)
               -> (ebind ref -> con)
               -> (convar * con -> convar * con)
               -> (kind -> kind)
               -> (module -> module)
               -> con -> con
    val map_exp : (expvar -> exp)
               -> (fbnd list -> fbnd list)
               -> (expvar * exp -> expvar * exp)
               -> (convar * exp -> convar * exp)
               -> (modvar * exp -> modvar * exp)
               -> (con -> con)
               -> (kind -> kind)
               -> (module -> module)
               -> exp -> exp
    val map_kind : (convar * kind -> convar * kind)
                -> (con -> con)
                -> kind -> kind
    val map_module : (modvar -> module)
                  -> (sbnd list -> sbnd list)
                  -> (modvar * module -> modvar * module)
                  -> (signat -> signat)
                  -> module -> module
    val map_sbnds : (expvar * sbnd list -> expvar * sbnd list)
                 -> (convar * sbnd list -> convar * sbnd list)
                 -> (modvar * sbnd list -> modvar * sbnd list)
                 -> (exp -> exp)
                 -> (con -> con)
                 -> (module -> module)
                 -> sbnd list -> sbnd list
    val map_sdecs : (expvar * sdec list -> expvar * sdec list)
                 -> (convar * sdec list -> convar * sdec list)
                 -> (modvar * sdec list -> modvar * sdec list)
                 -> (con -> con)
                 -> (kind -> kind)
                 -> (signat -> signat)
                 -> sdec list -> sdec list
    val map_signat : (modvar * signat -> modvar * signat)
                  -> (sdec list -> sdec list)
                  -> signat -> signat
  end
