structure IL1 :> IL1 =
  struct
    exception Error of string (* indicates internal fatal error *)

    type 'a var = 'a Variable.var

    (* the various "label" constructors correspond roughly to the overbar
       function in HS *)
    datatype label =
        LabTycon of string
      | LabTyvar of string
      | LabExp of string
      | LabModule of string
      | LabFunctor of string
      | LabRecord of string
      | LabSum of string
      | LabInternal of string (* for eq, expose, it, tag *)
      | LabStarred of label (* starred label *)

    fun streq x y =
        case String.compare (x, y) of
            EQUAL => true
          | _ => false

    fun label_eq (lab1, lab2) =
        case (lab1, lab2) of
            (LabTycon s1, LabTycon s2) => streq s1 s2
          | (LabTyvar s1, LabTyvar s2) => streq s1 s2
          | (LabExp s1, LabExp s2) => streq s1 s2
          | (LabModule s1, LabModule s2) => streq s1 s2
          | (LabFunctor s1, LabFunctor s2) => streq s1 s2
          | (LabRecord s1, LabRecord s2) => streq s1 s2
          | (LabSum s1, LabSum s2) => streq s1 s2
          | (LabInternal s1, LabInternal s2) => streq s1 s2
          | (LabStarred l1, LabStarred l2) => label_eq (l1, l2)
          | _ => false

    fun string_of_label label =
        case label of
            LabTycon s => s
          | LabTyvar s => s
          | LabExp s => s
          | LabModule s => s
          | LabFunctor s => s
          | LabRecord s => s
          | LabSum s => s
          | LabInternal s => s
          | LabStarred l => string_of_label l ^ "*"

    type convar = Variables.con var
    type modvar = Variables.mod var
    type expvar = Variables.exp var

    structure ConVarSet : ORD_SET =
        SplaySetFn (struct
                        type ord_key = convar
                        val compare = Variable.compare
                    end)

    structure ModVarSet : ORD_SET =
        SplaySetFn (struct
                        type ord_key = modvar
                        val compare = Variable.compare
                    end)

    datatype kind =
        KndType
      (* subsumes arrow *)
      | KndPi of convar * kind * kind (* convar bound in kind2 *)
      | KndProd of kind list
      | KndSing of con

    and con =
        ConVar of convar
      | ConEvar of ebind ref (* existential variable, for unification *)
      | ConAbs of convar * kind * con (* convar bound in con *)
      | ConApp of con * con
      | ConProd of con list
      | ConProj of int * con
      | ConMu of convar * kind * con (* convar bound in con, kind = KndType^n *)
      | ConModProj of module * label * label list
          (* invariant: module is a value *)

      (* cons with kind KndType *)
      | TyLit of Literal.lit_ty
      | TyArrow of con * con
      | TyRef of con
      | TyTagged
      | TyTag of con
      | TySum of (label * con) list
      | TyRecord of (label * con) list
      | TyForall of convar * kind * con (* convar bound in con *)

    (* Evar binding *)
    and ebind =
        EvFree of ConVarSet.set * ModVarSet.set * int
               (* con deps,       module deps,    id *)
               (* invariant: these variables are never
                  substituted into -- fatal error if they are *)
      | EvBound of con (* unified with con *)

    and exp =
        ExVar of expvar
      | ExLit of Literal.lit
      | ExUnOp of Literal.unop * exp
      | ExBinOp of exp * Literal.binop * exp
      | ExFix of fbnd list
      | ExApp of exp * exp
      | ExHandle of exp * exp
      | ExRaise of con * exp
      | ExRef of exp
      | ExGet of exp
      | ExSet of exp * exp
      | ExRoll of con * exp
      | ExUnroll of exp
      | ExRecord of (label * exp) list
      | ExRecordProj of label * exp * label list
      | ExSumInj of con * label * exp * label list
      | ExCase of con * exp * (label * exp) list (* con is result type *)
      | ExTagInj of exp * exp
      | ExNewTag of con
      | ExIfTagOf of exp * exp * exp * exp
      | ExConAbs of convar * kind * exp (* convar bound in exp *)
      | ExConApp of exp * con
      | ExLet of expvar * exp * exp (* expvar bound in exp2 *)
      | ExModLet of modvar * module * exp * con (* modvar bound in exp *)
      | ExModProj of module * label * label list

    and module =
        ModVar of modvar
      | ModStruct of (label * bnd) list
      | ModAbs of modvar * signat * module (* modvar bound in module *)
      | ModApp of module * module
      | ModModProj of module * label * label list
      | ModConstrain of module * signat
      | ModLet of modvar * module * module * signat
               (* modvar bound in module2 *)

    and bnd =
        BndExp of expvar * exp
      | BndCon of convar * con
      | BndMod of modvar * module
    (* NB: in a bnd list bnd::bnds, bnd's var bound in bnds *)

    and signat =
        SigStruct of (label * dec) list
      | SigPi of modvar * signat * signat (* modvar bound in signat2 *)

    and dec =
        DecExpCon of expvar * con
      | DecConKind of convar * kind
      | DecModSig of modvar * signat
    (* NB: in a dec list dec::decs, dec's var bound in decs *)

    withtype fbnd = expvar * expvar * con * con * exp
                   (* f      (x     : ty) : ty  = e *)
                   (* expvar2 (x) bound in exp (e),
                      expvar1 (f) bound in all exps *)

    type sdec = label * dec
    type sbnd = label * bnd

    (* polymorphic -- works on both sbnds and sdecs *)
    fun labdom sxxxs = map (fn (lab, _) => lab) sxxxs

    fun labelledKndSing con knd =
        case knd of
            KndType => KndSing con
          | KndSing _ => KndSing con
          | KndPi (cx, k1, k2) =>
                KndPi (cx, k1, labelledKndSing (ConApp (con, ConVar cx)) k2)
          | KndProd knds =>
                KndProd
                   (ListUtil.mapi
                        (fn (k, i) => labelledKndSing (ConProj (i+1, con)) k)
                        knds)


    (*** free module variables ***)

    fun fmv_con con =
        case con of
            ConVar _ => ModVarSet.empty
          | ConEvar ebr => ModVarSet.empty (* XXX should this be its moddeps? *)
          | ConAbs (cx, k, c) => fmv_con c
          | ConApp (c1, c2) => ModVarSet.union (fmv_con c1, fmv_con c2)
          | ConProd cs =>
                foldl (fn (c, set) => ModVarSet.union (fmv_con c, set))
                      ModVarSet.empty
                      cs
          | ConProj (i, c) => fmv_con c
          | ConMu (cx, k, c) => ModVarSet.union (fmv_kind k, fmv_con c)
          | ConModProj (module, lab, labs) => fmv_mod module
          | TyLit _ => ModVarSet.empty
          | TyArrow (c1, c2) => ModVarSet.union (fmv_con c1, fmv_con c2)
          | TyRef c => fmv_con c
          | TyTagged => ModVarSet.empty
          | TyTag c => fmv_con c
          | TySum labcs =>
                foldl (fn ((l, c), set) => ModVarSet.union (fmv_con c, set))
                      ModVarSet.empty
                      labcs
          | TyRecord labcs =>
                foldl (fn ((l, c), set) => ModVarSet.union (fmv_con c, set))
                      ModVarSet.empty
                      labcs
          | TyForall (cx, k, c) => ModVarSet.union (fmv_kind k, fmv_con c)

    and fmv_kind knd =
        case knd of
            KndType => ModVarSet.empty
          | KndPi (cx, k1, k2) => ModVarSet.union (fmv_kind k1, fmv_kind k2)
          | KndProd ks =>
                foldl (fn (k, set) => ModVarSet.union (fmv_kind k, set))
                      ModVarSet.empty
                      ks
          | KndSing c => fmv_con c

    and fmv_mod module =
        case module of
            ModVar mx => ModVarSet.singleton mx
          | ModStruct labbnds =>
                foldl (fn ((lab, bnd), set) =>
                        ModVarSet.union (fmv_bnd bnd, set))
                      ModVarSet.empty
                      labbnds
          | ModAbs (mx, sg, m) =>
                ModVarSet.union (fmv_sig sg, ModVarSet.delete (fmv_mod m, mx))
          | ModApp (m1, m2) => ModVarSet.union (fmv_mod m1, fmv_mod m2)
          | ModModProj (m, lab, labs) => fmv_mod m
          | ModConstrain (m, sg) => ModVarSet.union (fmv_mod m, fmv_sig sg)
          | ModLet (mx, m, m', sg') =>
                (* NB: mx not bound in sg', since (a) it makes map_module
                   easier to write and use, and (b) mx can't appear in sg'
                   *anyway*, for the ModLet to be well-formed *)
                ModVarSet.union
                   (fmv_mod m,
                    ModVarSet.union
                        (ModVarSet.delete (fmv_mod m', mx),
                         fmv_sig sg'))

    and fmv_bnd bnd =
        case bnd of
            BndExp (ex, e) => fmv_exp e
          | BndCon (cx, c) => fmv_con c
          | BndMod (mx, m) => ModVarSet.delete (fmv_mod m, mx)

    and fmv_sig signat =
        case signat of
            SigStruct sdecs =>
                foldl (fn ((lab, dec), set) =>
                        ModVarSet.union (fmv_dec dec, set))
                      ModVarSet.empty
                      sdecs
          | SigPi (mx, sg, sg') =>
                ModVarSet.union (fmv_sig sg, ModVarSet.delete (fmv_sig sg', mx))

    and fmv_dec dec =
        case dec of
            DecExpCon (ex, c) => fmv_con c
          | DecConKind (cx, k) => fmv_kind k
          | DecModSig (mx, sg) => ModVarSet.delete (fmv_sig sg, mx)
    

    and fmv_exp exp =
        case exp of
            ExVar _ => ModVarSet.empty
          | ExLit _ => ModVarSet.empty
          | ExUnOp (_, e) => fmv_exp e
          | ExBinOp (e1, _, e2) => ModVarSet.union(fmv_exp e1, fmv_exp e2)
          | ExFix fbnds =>
                foldl (fn ((f, x, c1, c2, e), set) =>
                            ModVarSet.union (fmv_exp e, set))
                      ModVarSet.empty
                      fbnds
          | ExApp (e1, e2) => ModVarSet.union (fmv_exp e1, fmv_exp e2)
          | ExRecord labes =>
                foldl (fn ((l, e), set) => ModVarSet.union (fmv_exp e, set))
                      ModVarSet.empty
                      labes
          | ExRecordProj (lab, e, labs) => fmv_exp e
          | ExHandle (e1, e2) => ModVarSet.union (fmv_exp e1, fmv_exp e2)
          | ExRaise (c, e) => ModVarSet.union (fmv_con c, fmv_exp e)
          | ExRef e => fmv_exp e
          | ExGet e => fmv_exp e
          | ExSet (e1, e2) => ModVarSet.union (fmv_exp e1, fmv_exp e2)
          | ExRoll (c, e) => ModVarSet.union (fmv_con c, fmv_exp e)
          | ExUnroll e => fmv_exp e
          | ExSumInj (c, lab, e, labs) => ModVarSet.union (fmv_con c, fmv_exp e)
          | ExCase (c, e, labes) =>
                ModVarSet.union
                    (fmv_con c,
                     ModVarSet.union
                        (fmv_exp e,
                         foldl (fn ((l, e), set) =>
                                    ModVarSet.union (fmv_exp e, set))
                               ModVarSet.empty
                               labes))
          | ExTagInj (e1, e2) => ModVarSet.union (fmv_exp e1, fmv_exp e2)
          | ExNewTag c => fmv_con c
          | ExIfTagOf (e1, e2, e3, e4) =>
                foldl (fn (e, set) => ModVarSet.union (fmv_exp e, set))
                      ModVarSet.empty
                      [e1, e2, e3, e4]
          | ExConAbs (cx, k, e) => ModVarSet.union (fmv_kind k, fmv_exp e)
          | ExConApp (e, c) => ModVarSet.union (fmv_exp e, fmv_con c)
          | ExLet (x, e1, e2) => ModVarSet.union (fmv_exp e1, fmv_exp e2)
          | ExModLet (mx, m, e, c) =>
                ModVarSet.union
                   (ModVarSet.union (fmv_mod m,
                                     ModVarSet.delete (fmv_exp e, mx)),
                    fmv_con c)
          | ExModProj (m, lab, labs) => fmv_mod m


    (*** map functions ***)

    (* map_kind : (convar * kind -> convar * kind) -> (con -> con)
               -> kind -> kind  *)
    fun map_kind fconvarkind fcon knd =
        let val fkind = map_kind fconvarkind fcon
        in
            case knd of
                KndType => knd
              | KndProd knds => KndProd (map fkind knds)
              | KndSing c => KndSing (fcon c)
              | KndPi (x, k1, k2) =>
                    let val k1 = fkind k1
                        val (x, k2) = fconvarkind (x, k2)
                    in
                        KndPi (x, k1, k2)
                    end
        end

    (* map_con : (convar -> con) -> (ebind ref -> con)
              -> (convar * con -> convar * con)
              -> (kind -> kind) -> (module -> module) -> con -> con *)
    (* XXX put these args in a record?  and for other map_* functions ... *)
    fun map_con fconvar febindref fconvarcon fkind fmodule con =
            (* XXX jooc, better to do "fun fcon con = ... con"
                   or  "val fcon = ..." ?  optimization of partial appl.? *)
        let fun fcon con = map_con fconvar febindref fconvarcon fkind fmodule con
            fun fconvarkindcon (x, k, c) =
                    let val k = fkind k
                        val (x, c) = fconvarcon (x, c)
                    in
                        (x, k, c)
                    end
            fun flabcons labcons = map (fn (lab, c) => (lab, fcon c)) labcons
        in
            case con of
                ConVar x => fconvar x
              | ConEvar ebref => febindref ebref
              | ConAbs xkc => ConAbs (fconvarkindcon xkc)
              | ConApp (c1, c2) => ConApp (fcon c1, fcon c2)
              | ConProd cons => ConProd (map fcon cons)
              | ConProj (i, c) => ConProj (i, fcon c)
              | ConMu xkc => ConMu (fconvarkindcon xkc)
              | ConModProj (m, lab, labs) => ConModProj (fmodule m, lab, labs)
              | TyLit lt => con
              | TyArrow (c1, c2) => TyArrow (fcon c1, fcon c2)
              | TyRef c => TyRef (fcon c)
              | TyTagged => con
              | TyTag c => TyTag (fcon c)
              | TySum labcons => TySum (flabcons labcons)
              | TyRecord labcons => TyRecord (flabcons labcons)
              | TyForall xkc => TyForall (fconvarkindcon xkc)
        end

    (* map_module : (modvar -> module)
                 -> ((label * bnd) list -> (label * bnd) list)
                 -> (modvar * module -> modvar * module)
                 -> (signat -> signat)
                 -> module -> module *)
    fun map_module fmodvar flabbnds fmodvarmodule fsignat module =
        let fun fmodule module = map_module fmodvar flabbnds fmodvarmodule
                                            fsignat module
            fun fmodvarsignatmodule (x, sg, m) =
                    let val sg = fsignat sg
                        val (x, m) = fmodvarmodule (x, m)
                    in
                        (x, sg, m)
                    end
            fun fmodlet (mx, m, m', sg') =
                    let val m = fmodule m
                        val sg' = fsignat sg'
                        val (mx, m') = fmodvarmodule (mx, m')
                    in
                        (mx, m, m', sg')
                    end
        in
            case module of
                ModVar mv => fmodvar mv
              | ModStruct labbnds => ModStruct (flabbnds labbnds)
              | ModAbs mv_sg_m => ModAbs (fmodvarsignatmodule mv_sg_m)
              | ModApp (m1, m2) => ModApp (fmodule m1, fmodule m2)
              | ModModProj (m, lab, labs) => ModModProj (fmodule m, lab, labs)
              | ModConstrain (m, sg) => ModConstrain (fmodule m, fsignat sg)
              | ModLet (mx_m_m'_sg') => ModLet (fmodlet (mx_m_m'_sg'))
        end

    (* map_sbnds : (expvar * sbnds -> expvar * sbnds)
                -> (convar * sbnds -> convar * sbnds)
                -> (modvar * sbnds -> modvar * sbnds)
                -> (exp -> exp) -> (con -> con) -> (module -> module)
                -> sbnds -> sbnds
       (where sbnds = (label * bnd) list) *)
    fun map_sbnds fexpvarsbnds fconvarsbnds fmodvarsbnds
                  fexp fcon fmodule sbnds =
        case sbnds of
            [] => []
          | (lab, b)::sbs =>
                case b of
                    BndExp (ex, e) =>
                        let val e = fexp e
                            val (ex, sbs) = fexpvarsbnds (ex, sbs)
                        in
                            (lab, BndExp (ex, e)) :: sbs
                        end
                  | BndCon (cx, c) =>
                        let val c = fcon c
                            val (cx, sbs) = fconvarsbnds (cx, sbs)
                        in
                            (lab, BndCon (cx, c)) :: sbs
                        end
                  | BndMod (mx, m) =>
                        let val m = fmodule m
                            val (mx, sbs) = fmodvarsbnds (mx, sbs)
                        in
                            (lab, BndMod (mx, m)) :: sbs
                        end

    (* map_signat : (modvar * signat -> modvar * signat)
                 -> ((lab * dec) list -> (lab * dec) list)
                 -> signat -> signat *)
    fun map_signat fmodvarsignat fsdecs signat =
        let fun fsignat signat = map_signat fmodvarsignat fsdecs signat
        in
            case signat of
                SigStruct sdecs => SigStruct (fsdecs sdecs)
              | SigPi (mx, sg1, sg2) =>
                    let val sg1 = fsignat sg1
                        val (mx, sg2) = fmodvarsignat (mx, sg2)
                    in
                        SigPi (mx, sg1, sg2)
                    end
        end

    (* map_sdecs : (expvar * sdecs -> expvar * sdecs)
                -> (convar * sdecs -> convar * sdecs)
                -> (modvar * sdecs -> modvar * sdecs)
                -> (con -> con) -> (kind -> kind) -> (signat -> signat)
                -> sdecs -> sdecs
       (where sdecs = (label * dec) list) *)
    fun map_sdecs fexpvarsdecs fconvarsdecs fmodvarsdecs
                  fcon fkind fsignat sdecs =
        case sdecs of
            [] => []
          | (lab, d)::sds =>
                case d of
                    DecExpCon (ex, c) =>
                        let val c = fcon c
                            val (ex, sds) = fexpvarsdecs (ex, sds)
                        in
                            (lab, DecExpCon (ex, c)) :: sds
                        end
                  | DecConKind (cx, k) =>
                        let val k = fkind k
                            val (cx, sds) = fconvarsdecs (cx, sds)
                        in
                            (lab, DecConKind (cx, k)) :: sds
                        end
                  | DecModSig (mx, sg) =>
                        let val sg = fsignat sg
                            val (mx, sds) = fmodvarsdecs (mx, sds)
                        in
                            (lab, DecModSig (mx, sg)) :: sds
                        end

    (* map_exp : (expvar -> exp)
              -> (fbnd list -> fbnd list)
              -> (expvar * exp -> expvar * exp)
              -> (convar * exp -> convar * exp)
              -> (modvar * exp -> modvar * exp)
              -> (con -> con)
              -> (kind -> kind)
              -> (module -> module)
              -> exp -> exp *)
    fun map_exp fexpvar
                ffbnds
                fexpvarexp
                fconvarexp
                fmodvarexp
                fcon
                fkind
                fmodule
                exp =
        let val fexp = map_exp fexpvar ffbnds fexpvarexp fconvarexp fmodvarexp
                               fcon fkind fmodule
            fun flabexps labexps = map (fn (l, e) => (l, fexp e)) labexps
            fun fconvarkindexp (x, k, e) =
                let val k = fkind k
                    val (x, e) = fconvarexp (x, e)
                in
                    (x, k, e)
                end
            fun fmodvarmodexpcon (mx, m, e, c) =
                let val m = fmodule m
                    val c = fcon c
                    val (mx, e) = fmodvarexp (mx, e)
                in
                    (mx, m, e, c)
                end
            fun fexpvarexpexp (x, e1, e2) =
                let val e1 = fexp e1
                    val (x, e2) = fexpvarexp (x, e2)
                in
                    (x, e1, e2)
                end
        in
            case exp of
                ExLit _ => exp
              | ExUnOp(oper, e) => ExUnOp(oper, fexp e)
              | ExBinOp(e1, oper, e2) => ExBinOp(fexp e1, oper, fexp e2)
              | ExVar x => fexpvar x
              | ExFix fbnds => ExFix (ffbnds fbnds)
              | ExApp (e1, e2) => ExApp (fexp e1, fexp e2)
              | ExRecord labexps => ExRecord (flabexps labexps)
              | ExRecordProj (lab, e, labs) => ExRecordProj (lab, fexp e, labs)
              | ExHandle (e1, e2) => ExHandle (fexp e1, fexp e2)
              | ExRaise (c, e) => ExRaise (fcon c, fexp e)
              | ExRef e => ExRef (fexp e)
              | ExGet e => ExGet (fexp e)
              | ExSet (e1, e2) => ExSet (fexp e1, fexp e2)
              | ExRoll (c, e) => ExRoll (fcon c, fexp e)
              | ExUnroll e => ExUnroll (fexp e)
              | ExSumInj (c, lab, e, labs) => ExSumInj (fcon c, lab, fexp e, labs)
              | ExCase (c, e, labes) => ExCase (fcon c, fexp e, flabexps labes)
              | ExTagInj (e1, e2) => ExTagInj (fexp e1, fexp e2)
              | ExNewTag c => ExNewTag (fcon c)
              | ExIfTagOf (e1, e2, e3, e4) => ExIfTagOf (fexp e1, fexp e2,
                                                         fexp e3, fexp e4)
              | ExConAbs xke => ExConAbs (fconvarkindexp xke)
              | ExConApp (e, c) => ExConApp (fexp e, fcon c)
              | ExLet xe1e2 => ExLet (fexpvarexpexp xe1e2)
              | ExModLet mxmec => ExModLet (fmodvarmodexpcon mxmec)
              | ExModProj (m, lab, labs) => ExModProj (fmodule m, lab, labs)
        end


    (* check_nocondep_ebindref : convar -> (con -> con) -> con *)
    (* [check_nocondep_ebindref convar k ebref] either
        (a) checks that convar doesn't appear in condeps
            if !ebref = EvFree (condeps, moddeps, n) or
        (b) returns [k con],
            if !ebref = EvBound con
        used in replace_con_con and subst_con_con *)
    fun check_nocondep_ebindref convar k ebref =
        case !ebref of
            EvFree (condeps, moddeps, n) =>
                if ConVarSet.member (condeps, convar) then
                    raise Error "substitution into Evar con dependency"
                else
                    ConEvar ebref
          | EvBound con => k con

    (* check_nomoddep_ebindref : modvar -> (con -> con) -> con *)
    (* same as above, but checks the module variable dependency set *)
    fun check_nomoddep_ebindref modvar k ebref =
        case !ebref of
            EvFree (condeps, moddeps, n) =>
                if ModVarSet.member (moddeps, modvar) then
                    raise Error "substitution into Evar module dependency"
                else
                    ConEvar ebref
          | EvBound con => k con


    (*** [replace_class1_class2 var var' c2] replaces all occurrences of
         var as a class1 object with var' inside a class2 object c2.
         this *doesn't* do any alpha conversion, so it's important that
         they only ever be called with a fresh var'.

         (these functions are used to implement alpha conversion, both in
         alpha equality and in capture-avoiding substitution, defined below
         as subst_class1_class2.)

         (c.f. alpha_vary, alpha_vary2)
    *)


    (** replacement of convars **)

    fun replace_con_kind convar convar' knd =
        map_kind (fn (cx, knd) =>
                    if Variable.eq (cx, convar) then
                        (* no free occurrences of var in knd -- bound by cx *)
                        (cx, knd)
                    else
                        (* cx not eq convar, so convar might be free in knd *)
                        (cx, replace_con_kind convar convar' knd))
                 (fn con => replace_con_con convar convar' con)
                 knd

    and replace_con_con convar convar' con =
        map_con (fn cx =>
                    if Variable.eq (cx, convar) then
                        ConVar convar'
                    else
                        ConVar cx)
                (check_nocondep_ebindref convar (replace_con_con convar convar'))
                (fn (cx, c) =>
                    if Variable.eq (cx, convar) then
                        (cx, c)
                    else
                        (cx, replace_con_con convar convar' c))
                (replace_con_kind convar convar')
                (replace_con_module convar convar')
                con

    and replace_con_module convar convar' module =
        map_module (fn mv => ModVar mv)
                   (replace_con_sbnds convar convar')
                   (fn (x, m) => (x, replace_con_module convar convar' m))
                   (replace_con_signat convar convar')
                   module


    and replace_con_sbnds convar convar' sbnds =
        map_sbnds (fn (ex, sbs) => (ex, replace_con_sbnds convar convar' sbs))
                  (fn (cx, sbs) => if Variable.eq (cx, convar) then
                                    (cx, sbs)
                                   else
                                    (cx, replace_con_sbnds convar convar' sbs))
                  (fn (mx, sbs) => (mx, replace_con_sbnds convar convar' sbs))
                  (replace_con_exp convar convar')
                  (replace_con_con convar convar')
                  (replace_con_module convar convar')
                  sbnds

    and replace_con_signat convar convar' signat =
        map_signat (fn (mx, sg) => (mx, replace_con_signat convar convar' sg))
                   (replace_con_sdecs convar convar')
                   signat

    and replace_con_sdecs convar convar' sdecs =
        map_sdecs (fn (ex, sds) => (ex, replace_con_sdecs convar convar' sds))
                  (fn (cx, sds) => if Variable.eq (cx, convar) then
                                    (cx, sds)
                                   else
                                    (cx, replace_con_sdecs convar convar' sds))
                  (fn (mx, sds) => (mx, replace_con_sdecs convar convar' sds))
                  (replace_con_con convar convar')
                  (replace_con_kind convar convar')
                  (replace_con_signat convar convar')
                  sdecs

    and replace_con_exp convar convar' exp =
        map_exp (fn x => ExVar x)
                (fn fbnds =>
                    map (fn (f, x, c1, c2, e) =>
                            (f, x, replace_con_con convar convar' c1,
                                   replace_con_con convar convar' c2,
                                   replace_con_exp convar convar' e))
                        fbnds)
                (fn (ex, e) => (ex, replace_con_exp convar convar' e))
                (fn (cx, e) =>
                    if Variable.eq (cx, convar) then
                        (cx, e)
                    else
                        (cx, replace_con_exp convar convar' e))
                (fn (mx, e) => (mx, replace_con_exp convar convar' e))
                (replace_con_con convar convar')
                (replace_con_kind convar convar')
                (replace_con_module convar convar')
                exp


    (** replacement of modvars **)

    fun replace_mod_kind modvar modvar' knd =
        map_kind (fn (cx, knd) => (cx, replace_mod_kind modvar modvar' knd))
                 (fn con => replace_mod_con modvar modvar' con)
                 knd

    and replace_mod_con modvar modvar' con =
        map_con (fn cx => ConVar cx)
                (check_nomoddep_ebindref modvar (replace_mod_con modvar modvar'))
                (fn (cx, c) => (cx, replace_mod_con modvar modvar' c))
                (replace_mod_kind modvar modvar')
                (replace_mod_module modvar modvar')
                con

    and replace_mod_module modvar modvar' module =
        map_module (fn mx =>
                        if Variable.eq (mx, modvar) then
                            ModVar modvar'
                        else
                            ModVar mx)
                   (replace_mod_sbnds modvar modvar')
                   (fn (mx, m) =>
                        if Variable.eq (mx, modvar) then
                            (mx, m)
                        else
                            (mx, replace_mod_module modvar modvar' m))
                   (replace_mod_signat modvar modvar')
                   module


    and replace_mod_sbnds modvar modvar' sbnds =
        map_sbnds (fn (ex, sbs) => (ex, replace_mod_sbnds modvar modvar' sbs))
                  (fn (cx, sbs) => (cx, replace_mod_sbnds modvar modvar' sbs))
                  (fn (mx, sbs) => if Variable.eq (mx, modvar) then
                                    (mx, sbs)
                                   else
                                    (mx, replace_mod_sbnds modvar modvar' sbs))
                  (replace_mod_exp modvar modvar')
                  (replace_mod_con modvar modvar')
                  (replace_mod_module modvar modvar')
                  sbnds

    and replace_mod_signat modvar modvar' signat =
        map_signat (fn (mx, sg) =>
                        if Variable.eq (mx, modvar) then
                            (mx, sg)
                        else
                            (mx, replace_mod_signat modvar modvar' sg))
                   (replace_mod_sdecs modvar modvar')
                   signat

    and replace_mod_sdecs modvar modvar' sdecs =
        map_sdecs (fn (ex, sds) => (ex, replace_mod_sdecs modvar modvar' sds))
                  (fn (cx, sds) => (cx, replace_mod_sdecs modvar modvar' sds))
                  (fn (mx, sds) => if Variable.eq (mx, modvar) then
                                    (mx, sds)
                                   else
                                    (mx, replace_mod_sdecs modvar modvar' sds))
                  (replace_mod_con modvar modvar')
                  (replace_mod_kind modvar modvar')
                  (replace_mod_signat modvar modvar')
                  sdecs

    and replace_mod_exp modvar modvar' exp =
        map_exp (fn x => ExVar x)
                (fn fbnds =>
                    map (fn (f, x, c1, c2, e) =>
                            (f, x, replace_mod_con modvar modvar' c1,
                                   replace_mod_con modvar modvar' c2,
                                   replace_mod_exp modvar modvar' e))
                        fbnds)
                (fn (ex, e) => (ex, replace_mod_exp modvar modvar' e))
                (fn (cx, e) => (cx, replace_mod_exp modvar modvar' e))
                (fn (mx, e) =>
                    if Variable.eq (mx, modvar) then
                        (mx, e)
                    else
                        (mx, replace_mod_exp modvar modvar' e))
                (replace_mod_con modvar modvar')
                (replace_mod_kind modvar modvar')
                (replace_mod_module modvar modvar')
                exp


    (** replacement of expvars **)

    fun replace_exp_kind expvar expvar' knd =
        map_kind (fn (cx, knd) => (cx, replace_exp_kind expvar expvar' knd))
                 (fn con => replace_exp_con expvar expvar' con)
                 knd

    and replace_exp_con expvar expvar' con =
        map_con (fn cx => ConVar cx)
                (fn ebr => case !ebr of
                                EvFree _ => ConEvar ebr
                              | EvBound c => replace_exp_con expvar expvar' c)
                (fn (cx, c) => (cx, replace_exp_con expvar expvar' c))
                (replace_exp_kind expvar expvar')
                (replace_exp_module expvar expvar')
                con

    and replace_exp_module expvar expvar' module =
        map_module (fn mx => ModVar mx)
                   (replace_exp_sbnds expvar expvar')
                   (fn (mx, m) => (mx, replace_exp_module expvar expvar' m))
                   (replace_exp_signat expvar expvar')
                   module

    and replace_exp_sbnds expvar expvar' sbnds =
        map_sbnds 
                  (fn (ex, sbs) => if Variable.eq (ex, expvar) then
                                    (ex, sbs)
                                   else
                                    (ex, replace_exp_sbnds expvar expvar' sbs))
                  (fn (cx, sbs) => (cx, replace_exp_sbnds expvar expvar' sbs))
                  (fn (mx, sbs) => (mx, replace_exp_sbnds expvar expvar' sbs))
                  (replace_exp_exp expvar expvar')
                  (replace_exp_con expvar expvar')
                  (replace_exp_module expvar expvar')
                  sbnds

    and replace_exp_signat expvar expvar' signat =
        map_signat (fn (mx, sg) => (mx, replace_exp_signat expvar expvar' sg))
                   (replace_exp_sdecs expvar expvar')
                   signat

    and replace_exp_sdecs expvar expvar' sdecs =
        map_sdecs (fn (ex, sds) => if Variable.eq (ex, expvar) then
                                    (ex, sds)
                                   else
                                    (ex, replace_exp_sdecs expvar expvar' sds))
                  (fn (cx, sds) => (cx, replace_exp_sdecs expvar expvar' sds))
                  (fn (mx, sds) => (mx, replace_exp_sdecs expvar expvar' sds))
                  (replace_exp_con expvar expvar')
                  (replace_exp_kind expvar expvar')
                  (replace_exp_signat expvar expvar')
                  sdecs

    and replace_exp_exp expvar expvar' exp =
        map_exp (fn x =>
                    if Variable.eq (x, expvar) then
                        ExVar expvar'
                    else
                        ExVar x)
                (fn fbnds =>
                    if List.exists
                        (fn (f, _, _, _, _) => Variable.eq (f, expvar))
                        fbnds
                    then
                        (* expvar is bound in entire phrase, do nothing *)
                        fbnds
                    else
                        map (fn (f, x, c1, c2, e) =>
                                let val e =
                                    if Variable.eq (x, expvar) then
                                        (* expvar is bound in e, do nothing *)
                                        e
                                    else
                                        replace_exp_exp expvar expvar' e
                                in

                                    (f, x, replace_exp_con expvar expvar' c1,
                                           replace_exp_con expvar expvar' c2,
                                           e)
                                end)
                            fbnds)
                (fn (ex, e) =>
                    if Variable.eq (ex, expvar) then
                        (ex, e)
                    else
                        (ex, replace_exp_exp expvar expvar' e))
                (fn (cx, e) => (cx, replace_exp_exp expvar expvar' e))
                (fn (mx, e) => (mx, replace_exp_exp expvar expvar' e))
                (replace_exp_con expvar expvar')
                (replace_exp_kind expvar expvar')
                (replace_exp_module expvar expvar')
                exp


    (*** [subst_class1_class1 class1 class1var class2] performs capture-
         avoiding substitution of class1 for class1var in class2 *)

    (* [alpha_vary (c1var, c2obj) replace] alpha-varies class1 bound
       variable c1var to a fresh variable in class2 object c2obj.
       replace replaces a class1 variable in a class2 object. *)
    fun alpha_vary (c1var, c2obj) replace =
        let val c1var' = Variable.freshen c1var
            val c2obj' = replace c1var c1var' c2obj
        in
            (c1var', c2obj')
        end

    (* alpha_vary2 : ('a var * 'b) -> ('a var * 'b)
                  -> ('a var -> 'a var -> 'b -> 'b)
                  -> ('a var * ('b * 'b))

       varies two binding constructs to use the same (fresh) variable.
       for computing alpha equality *)
    fun alpha_vary2 (varA, objA) (varB, objB) replace =
        let val var' = Variable.freshen varA
            val objA' = replace varA var' objA
            val objB' = replace varB var' objB
        in
            (var', (objA', objB'))
        end

    (* alpha_vary_many : 'a var list * 'b list
                      -> ('a var -> 'a var -> 'b -> 'b)
                      -> 'a var list * 'b list *)
    (* varies many bound variables in many objects. 
       used in subst_exp_exp, for fbnds substitution.*)
    (* not the most efficient -- would be better to do a multiary replace... *)
    fun alpha_vary_many (vars, objs) replace =
        let val vars' = map Variable.freshen vars
            fun subst obj = ListPair.foldl (fn (v, v', ob) => replace v v' ob)
                                           obj
                                           (vars, vars')
            val objs' = map subst objs
        in
            (vars', objs')
        end

    (* special function for alpha-varying a dec binding into an sdecs *)
    fun alpha_vary_dec_sdecs (dec, sdecs) =
        case dec of
            DecExpCon (x, c) =>
                let val (x, sdecs) = alpha_vary (x, sdecs) replace_exp_sdecs
                in
                    (DecExpCon (x, c), sdecs)
                end
          | DecConKind (cx, k) =>
                let val (cx, sdecs) = alpha_vary (cx, sdecs) replace_con_sdecs
                in
                    (DecConKind (cx, k), sdecs)
                end
          | DecModSig (mx, sg) =>
                let val (mx, sdecs) = alpha_vary (mx, sdecs) replace_mod_sdecs
                in
                    (DecModSig (mx, sg), sdecs)
                end

    (* special function for alpha-varying a bnd binding into an sbnds *)
    fun alpha_vary_bnd_sbnds (bnd, sbnds) =
        case bnd of
            BndExp (x, e) =>
                let val (x, sbnds) = alpha_vary (x, sbnds) replace_exp_sbnds
                in
                    (BndExp (x, e), sbnds)
                end
          | BndCon (cx, c) =>
                let val (cx, sbnds) = alpha_vary (cx, sbnds) replace_con_sbnds
                in
                    (BndCon (cx, c), sbnds)
                end
          | BndMod (mx, m) =>
                let val (mx, sbnds) = alpha_vary (mx, sbnds) replace_mod_sbnds
                in
                    (BndMod (mx, m), sbnds)
                end


    (* constructor substitution *)

    fun subst_con_kind con convar knd =
        map_kind (fn (cx, k) =>
                    if Variable.eq (cx, convar) then
                        (cx, k)
                    else
                        let (* eagerly alpha vary with a fresh variable *)
                            val (cx, k) = alpha_vary (cx, k) replace_con_kind
                            (* cx \not\in fv(con), so we won't incur capture *)
                            val k = subst_con_kind con convar k
                        in
                            (cx, k)
                        end)
                 (subst_con_con con convar)
                 knd

    and subst_con_con confor convar incon =
        map_con (fn cx =>
                    if Variable.eq (cx, convar) then
                        (* cx is convar free, substitute for it *)
                        confor
                    else
                        (* not convar -- put it back *)
                        ConVar cx)
                (check_nocondep_ebindref convar (subst_con_con confor convar))
                (fn (cx, c) =>
                    if Variable.eq (cx, convar) then
                        (* cx is convar being rebound, leave it alone *)
                        (cx, c)
                    else
                        (* different variable, eagerly rename and continue *)
                        let val (cx, c) = alpha_vary (cx, c) replace_con_con
                            (* now, cx \not\in fv(confor) *)
                            val c = subst_con_con confor convar c
                        in
                            (cx, c)
                        end)
                (subst_con_kind confor convar)
                (subst_con_module confor convar)
                incon

    and subst_con_module con convar module =
        map_module (fn mx => ModVar mx)
                   (subst_con_sbnds con convar)
                   (fn (mx, m) => (mx, subst_con_module con convar m))
                   (subst_con_signat con convar)
                   module

    and subst_con_sbnds con convar sbnds =
        map_sbnds (fn (ex, sbs) => (ex, subst_con_sbnds con convar sbs))
                  (fn (cx, sbs) => if Variable.eq (cx, convar) then   
                                    (cx, sbs)
                                   else
                                    let val (cx, sbs) = alpha_vary (cx, sbs)
                                                            replace_con_sbnds
                                        val sbs = subst_con_sbnds con convar sbs
                                    in
                                        (cx, sbs)
                                    end)
                  (fn (mx, sbs) => (mx, subst_con_sbnds con convar sbs))
                  (subst_con_exp con convar)
                  (subst_con_con con convar)
                  (subst_con_module con convar)
                  sbnds

    and subst_con_signat con convar signat =
        map_signat (fn (mx, sg) => (mx, subst_con_signat con convar sg))
                   (subst_con_sdecs con convar)
                   signat

    and subst_con_sdecs con convar sdecs =
        map_sdecs (fn (ex, sds) => (ex, subst_con_sdecs con convar sds))
                  (fn (cx, sds) => if Variable.eq (cx, convar) then
                                    (cx, sds)
                                   else
                                    let val (cx, sds) = alpha_vary (cx, sds)
                                                            replace_con_sdecs
                                        val sds = subst_con_sdecs con convar sds
                                    in
                                        (cx, sds)
                                    end)
                  (fn (mx, sds) => (mx, subst_con_sdecs con convar sds))
                  (subst_con_con con convar)
                  (subst_con_kind con convar)
                  (subst_con_signat con convar)
                  sdecs

    and subst_con_exp con convar exp =
        map_exp (fn x => ExVar x)
                (fn fbnds =>
                    map (fn (f, x, c1, c2, e) =>
                            (f, x, subst_con_con con convar c1,
                                   subst_con_con con convar c2,
                                   subst_con_exp con convar e))
                        fbnds)
                (fn (ex, e) => (ex, subst_con_exp con convar e))
                (fn (cx, e) =>
                    if Variable.eq (cx, convar) then
                        (cx, e)
                    else
                        let val (cx, e) = alpha_vary (cx, e) replace_con_exp
                            val e = subst_con_exp con convar e
                        in
                            (cx, e)
                        end)
                (fn (mx, e) => (mx, subst_con_exp con convar e))
                (subst_con_con con convar)
                (subst_con_kind con convar)
                (subst_con_module con convar)
                exp


    (* module substitution *)

    fun subst_mod_kind module modvar knd =
        map_kind (fn (cx, k) => (cx, subst_mod_kind module modvar k))
                 (subst_mod_con module modvar)
                 knd

    and subst_mod_con module modvar con =
        map_con (fn cx => ConVar cx)
                (check_nomoddep_ebindref modvar (subst_mod_con module modvar))
                (fn (cx, c) => (cx, subst_mod_con module modvar c))
                (subst_mod_kind module modvar)
                (subst_mod_module module modvar)
                con

    and subst_mod_module modulefor modvar modulein =
        map_module (fn mx =>
                        if Variable.eq (mx, modvar) then
                            modulefor
                        else
                            ModVar mx)
                   (subst_mod_sbnds modulefor modvar)
                   (fn (mx, m) =>
                        if Variable.eq (mx, modvar) then
                            (mx, m)
                        else
                            let val (mx, m) = alpha_vary (mx, m)
                                                         replace_mod_module
                                val m = subst_mod_module modulefor modvar m
                            in
                                (mx, m)
                            end)
                   (subst_mod_signat modulefor modvar)
                   modulein

    and subst_mod_sbnds module modvar sbnds =
        map_sbnds (fn (ex, sbs) => (ex, subst_mod_sbnds module modvar sbs))
                  (fn (cx, sbs) => (cx, subst_mod_sbnds module modvar sbs))
                  (fn (mx, sbs) =>
                        if Variable.eq (mx, modvar) then   
                            (mx, sbs)
                        else
                            let val (mx, sbs) = alpha_vary (mx, sbs)
                                                           replace_mod_sbnds
                                val sbs = subst_mod_sbnds module modvar sbs
                            in
                                (mx, sbs)
                            end)
                  (subst_mod_exp module modvar)
                  (subst_mod_con module modvar)
                  (subst_mod_module module modvar)
                  sbnds

    and subst_mod_signat module modvar signat =
        map_signat (fn (mx, sg) =>
                        if Variable.eq (mx, modvar) then
                            (mx, sg)
                        else
                            let val (mx, sg) = alpha_vary (mx, sg)
                                                          replace_mod_signat
                                val sg = subst_mod_signat module modvar sg
                            in
                                (mx, sg)
                            end)
                   (subst_mod_sdecs module modvar)
                   signat

    and subst_mod_sdecs module modvar sdecs =
        map_sdecs (fn (ex, sds) => (ex, subst_mod_sdecs module modvar sds))
                  (fn (cx, sds) => (cx, subst_mod_sdecs module modvar sds))
                  (fn (mx, sds) =>
                        if Variable.eq (mx, modvar) then
                            (mx, sds)
                        else
                            let val (mx, sds) = alpha_vary (mx, sds)
                                                           replace_mod_sdecs
                                val sds = subst_mod_sdecs module modvar sds
                            in
                                (mx, sds)
                            end)
                  (subst_mod_con module modvar)
                  (subst_mod_kind module modvar)
                  (subst_mod_signat module modvar)
                  sdecs

    and subst_mod_exp module modvar exp =
        map_exp (fn x => ExVar x)
                (fn fbnds =>
                    map (fn (f, x, c1, c2, e) =>
                            (f, x, subst_mod_con module modvar c1,
                                   subst_mod_con module modvar c2,
                                   subst_mod_exp module modvar e))
                        fbnds)
                (fn (ex, e) => (ex, subst_mod_exp module modvar e))
                (fn (cx, e) => (cx, subst_mod_exp module modvar e))
                (fn (mx, e) =>
                    if Variable.eq (mx, modvar) then
                        (mx, e)
                    else
                        let val (mx, e) = alpha_vary (mx, e) replace_mod_exp
                            val e = subst_mod_exp module modvar e
                        in
                            (mx, e)
                        end)
                (subst_mod_con module modvar)
                (subst_mod_kind module modvar)
                (subst_mod_module module modvar)
                exp


    (* expression substitution *)

    fun subst_exp_kind exp expvar knd =
        map_kind (fn (cx, k) => (cx, subst_exp_kind exp expvar k))
                 (subst_exp_con exp expvar)
                 knd

    and subst_exp_con exp expvar con =
        map_con (fn cx => ConVar cx)
                (fn ebr => case !ebr of
                                EvFree _ => ConEvar ebr
                              | EvBound c => subst_exp_con exp expvar c)
                (fn (cx, c) => (cx, subst_exp_con exp expvar c))
                (subst_exp_kind exp expvar)
                (subst_exp_module exp expvar)
                con

    and subst_exp_module exp expvar module =
        map_module (fn mx => ModVar mx)
                   (subst_exp_sbnds exp expvar)
                   (fn (mx, m) => (mx, subst_exp_module exp expvar m))
                   (subst_exp_signat exp expvar)
                   module

    and subst_exp_sbnds exp expvar sbnds =
        map_sbnds (fn (ex, sbs) =>
                        if Variable.eq (ex, expvar) then   
                            (ex, sbs)
                        else
                            let val (ex, sbs) = alpha_vary (ex, sbs)
                                                           replace_exp_sbnds
                                val sbs = subst_exp_sbnds exp expvar sbs
                            in
                                (ex, sbs)
                            end)
                  (fn (cx, sbs) => (cx, subst_exp_sbnds exp expvar sbs))
                  (fn (mx, sbs) => (mx, subst_exp_sbnds exp expvar sbs))
                  (subst_exp_exp exp expvar)
                  (subst_exp_con exp expvar)
                  (subst_exp_module exp expvar)
                  sbnds

    and subst_exp_signat exp expvar signat =
        map_signat (fn (mx, sg) => (mx, subst_exp_signat exp expvar signat))
                   (subst_exp_sdecs exp expvar)
                   signat

    and subst_exp_sdecs exp expvar sdecs =
        map_sdecs (fn (ex, sds) =>
                        if Variable.eq (ex, expvar) then
                            (ex, sds)
                        else
                            let val (ex, sds) = alpha_vary (ex, sds)
                                                           replace_exp_sdecs
                                val sds = subst_exp_sdecs exp expvar sds
                            in
                                (ex, sds)
                            end)
                  (fn (cx, sds) => (cx, subst_exp_sdecs exp expvar sds))
                  (fn (mx, sds) => (mx, subst_exp_sdecs exp expvar sds))
                  (subst_exp_con exp expvar)
                  (subst_exp_kind exp expvar)
                  (subst_exp_signat exp expvar)
                  sdecs

    and subst_exp_exp expfor expvar expin =
        map_exp (fn x =>
                    if Variable.eq (x, expvar) then
                        expfor
                    else
                        ExVar x)
                (fn fbnds =>
                    if List.exists
                            (fn (f, _, _, _, _) => Variable.eq (f, expvar))
                            fbnds
                    then
                        (* expvar is bound in entire phrase, do nothing *)
                        fbnds
                    else
                        let (* alpha-vary all the function variables *)
                            val fes = map (fn (f, _, _, _, e) => (f, e)) fbnds
                            val (fs, es) = ListPair.unzip fes
                            val (fs, es) = alpha_vary_many (fs, es)
                                           replace_exp_exp
                            val fes = ListPair.zip (fs, es)
                            val fbnds = ListPair.map
                                            (fn ((f, e), (_, x, c1, c2, _)) =>
                                                (f, x, c1, c2, e))
                                            (fes, fbnds)
                        in
                            map (fn (f, x, c1, c2, e) =>
                                  let val (x, e) =
                                      if Variable.eq (x, expvar) then
                                          (x, e)
                                      else
                                          let val (x, e) = alpha_vary (x, e)
                                                           replace_exp_exp
                                              val e = subst_exp_exp
                                                      expfor expvar e
                                          in
                                              (x, e)
                                          end
                                  in
                                      (f, x, subst_exp_con expfor expvar c1,
                                             subst_exp_con expfor expvar c2,
                                             e)
                                  end)
                                fbnds
                        end)
                (fn (ex, e) =>
                    if Variable.eq (ex, expvar) then
                        (ex, e)
                    else
                        let val (ex, e) = alpha_vary (ex, e) replace_exp_exp
                            val e = subst_exp_exp expfor expvar e
                        in
                            (ex, e)
                        end)
                (fn (cx, e) => (cx, subst_exp_exp expfor expvar e))
                (fn (ex, e) => (ex, subst_exp_exp expfor expvar e))
                (subst_exp_con expfor expvar)
                (subst_exp_kind expfor expvar)
                (subst_exp_module expfor expvar)
                expin

    
    (*** derived forms ***)

    (* NB: some elaborator-specific derived forms are in Elaborator *)
    fun derivedProd Con l =
        Con (ListUtil.mapi (fn (c, i) => (LabRecord (Int.toString (i+1)), c)) l)

    val tyProd = derivedProd TyRecord
    val exProd = derivedProd ExRecord
    fun exProj i exp max =
       ExRecordProj(LabRecord (Int.toString i), exp,
        List.tabulate (max, fn x => LabRecord (Int.toString (1 + x))))

    val tyUnit = TyRecord []
    val exUnit = ExRecord []

    val kndUnit = KndProd []
    val conUnit = ConProd []

    val tyBool = TySum [(LabSum "1", tyUnit), (LabSum "2", tyUnit)]
    val exFalse = ExSumInj (tyBool, LabSum "1", exUnit, [LabSum "1", LabSum "2"])
    val exTrue = ExSumInj (tyBool, LabSum "2", exUnit, [LabSum "1", LabSum "2"])
    fun exBool true = exTrue
      | exBool false = exFalse

    val tyInt = TyLit Literal.LitTyInt
    val tyChar = TyLit Literal.LitTyChar
    val tyString = TyLit Literal.LitTyString

    fun exInt n = ExLit (Literal.LitInt n)
    fun exChar c = ExLit (Literal.LitChar c)
    fun exString s = ExLit (Literal.LitString s)
    

    fun kndArrow (k1, k2) =
        let val cx = Variable.newNamed "_"
        in
            KndPi (cx, k1, k2)
        end

    fun sigArrow (sg1, sg2) =
        let val mx = Variable.newNamed "_"
        in
            SigPi (mx, sg1, sg2)
        end

    fun kndExp (k, n) = KndProd (List.tabulate (n, fn _ => k))

    fun exLam (var, con, con', exp) =
        let val var' = Variable.newNamed "_f"
        in
            ExRecordProj (LabRecord "1", ExFix [(var', var, con, con', exp)], [LabRecord "1"])
        end
    fun exPproj (labi, sumcon, exp, exn_exp) =
        let val labcons = case sumcon of
                            TySum lc => lc
                          | _ => raise Error "exPproj at non-sum type"
            val coni = case ListUtil.Alist.find label_eq labcons labi of
                            SOME con => con
                          | NONE => raise Error "exPproj on non-existent label"
            val ex : expvar = Variable.new ()
            val raise_exp = ExRaise (coni, exn_exp)
            val labes =
                map (fn (lab, con) =>
                        (lab, exLam (ex, con, coni, if label_eq (lab, labi) then
                                                     ExVar ex
                                                    else
                                                     raise_exp)))
                        labcons
        in
            ExCase (coni, exp, labes)
        end
    
    fun exLet (bnds, e) =
      let
        fun folder (BndExp(x, e'), e) =
             let val (x, e) = alpha_vary (x, e) replace_exp_exp
             in ExLet(x, e', e)
             end
         | folder _ = raise Error "exLet only takes expression bindings"
      in
        foldr folder e bnds
      end
      

    (* con is the result type *)
    fun exIf con (e1, e2, e3) =
        let val ex = Variable.newNamed "_"
        in
            ExCase (con, e1, [(LabSum "1", exLam (ex, tyUnit, con, e2)),
                              (LabSum "2", exLam (ex, tyUnit, con, e3))])
        end

    fun exPproj (labi, sumcon, exp, exn_exp) =
        let val labcons = case sumcon of
                            TySum lc => lc
                          | _ => raise Error "exPproj at non-sum type"
            val coni = case ListUtil.Alist.find label_eq labcons labi of
                            SOME con => con
                          | NONE => raise Error "exPproj on non-existent label"
            val ex : expvar = Variable.new ()
            val raise_exp = ExRaise (coni, exn_exp)
            val labes =
                map (fn (lab, con) =>
                        (lab, exLam (ex, con, coni, if label_eq (lab, labi) then
                                                     ExVar ex
                                                    else
                                                     raise_exp)))
                        labcons
        in
            ExCase (coni, exp, labes)
        end

    (* (a1 : T, ..., an : T) . c; for use in conMultiAbs and conMultiMu below *)
    fun multiBind (convars : convar list, phrase) subst_con_phrase =
        let val cx : convar = Variable.new ()
            val n = List.length convars
            val k = kndExp (KndType, n)
            (* substitution = [pi_i cx / a1, ..., pi_n cx / an] (i = 1..n) *)
            val substitution =
                ListUtil.mapi
                          (fn (cxi, i) => (ConProj (i+1, ConVar cx), cxi))
                          convars
            val phrase =
                foldr (fn ((cproji, cxi), c) => subst_con_phrase cproji cxi c)
                      phrase
                      substitution
        in
            (cx, k, phrase)
        end

    (* \ (a1 : T, ..., an : T) . c *)
    fun conMultiAbs (convars, con) =
        let val cx_k_con = multiBind (convars, con) subst_con_con
        in
            ConAbs cx_k_con
        end

    fun kndMultiPi (convars, knd) =
        let val cx_k_knd = multiBind (convars, knd) subst_con_kind
        in
            KndPi cx_k_knd
        end

    (* mu (a1 : T, ..., an : T) . c *)
    fun conMultiMu (convars, con) =
        let val cx_k_con = multiBind (convars, con) subst_con_con
        in
            ConMu cx_k_con
        end

    (* /\ (a1 : T, ..., an : T) . e *)
    fun exMultiConAbs (convars, exp) =
        let val cx_k_exp = multiBind (convars, exp) subst_con_exp
        in
            ExConAbs cx_k_exp
        end

    fun tyMultiForall (convars, con) =
        let val cx_k_con = multiBind (convars, con) subst_con_con
        in
            TyForall cx_k_con
        end
  end
