structure IL2 :> IL2 =
  struct
    type expvar = Variables.exp Variable.var
    type convar = Variables.con Variable.var
    type modvar = Variables.mod Variable.var

    datatype kind =
        KndType
      | KndPi of convar * kind * kind (* convar bound in kind2 *)
      | KndUnit
      | KndSigma of convar * kind * kind (* convar bound in kind2 *)
      | KndSing of con

    and con =
        ConVar of convar
      | ConAbs of convar * kind * con
      | ConApp of con * con
      | ConUnit
      | ConPair of con * con
      | ConProj1 of con
      | ConProj2 of con
      | ConMu of convar * kind * con (* convar bound in con, kind = KndType^n *)

      | TyLit of Literal.lit_ty
      | TyArrow of con * con
      | TyRef of con
      | TyTagged
      | TyTag of con
      | TySum of con list
      | TyProd of con list
      | TyForall of convar * kind * con (* convar bound in con *)
    
    datatype exp =
        ExVar of expvar
      | ExLit of Literal.lit
      | ExFix of fbnd list
      | ExLam of expvar * con * exp
      | ExUnOp of Literal.unop * exp
      | ExBinOp of exp * Literal.binop * exp
      | ExApp of exp * exp
      | ExHandle of exp * exp
      | ExRaise of con * exp
      | ExRef of exp
      | ExGet of exp
      | ExSet of exp * exp
      | ExRoll of con * exp
      | ExUnroll of exp
      | ExProd of exp list
      | ExProj of int * exp
      | ExSumInj of con * int * exp
      | ExCase of con * exp * exp list
      | ExTagInj of exp * exp
      | ExNewTag of con
      | ExIfTagOf of exp * exp * exp * exp
      | ExConAbs of convar * kind * exp (* convar bound in exp *)
      | ExConApp of exp * con
      | ExLet of expvar * exp * exp (* expvar bound in exp2 *)

    (* degenerate module calculus *)
    and module = Mod of con * exp
    and signat = Sig of convar * kind * con (* convar bound in con *)

    withtype fbnd = (expvar * expvar * con * con * exp)


    (*** derived forms ***)

    fun kndPair (k1, k2) =
        let val cx = Variable.newNamed "_"
        in
            KndSigma (cx, k1, k2)
        end

    fun kndProd ks = foldr kndPair KndUnit ks

    fun conProd cs = foldr ConPair ConUnit cs

    fun conProj (n, c) =
        if n = 1
        then ConProj1 c
        else conProj (n-1, ConProj2 c)

    val tyUnit = TyProd []
    val tyBool = TySum [tyUnit, tyUnit]
    val tyInt = TyLit Literal.LitTyInt
    val tyChar = TyLit Literal.LitTyChar
    val tyString = TyLit Literal.LitTyString

    val exUnit = ExProd []
    fun exBool b = ExSumInj (tyBool, if b then 1 else 2, exUnit)
    fun exInt n = ExLit (Literal.LitInt n)
    fun exChar c = ExLit (Literal.LitChar c)
    fun exString s = ExLit (Literal.LitString s)

    (*** mapping ***)

    (* map_kind : (convar * kind -> convar * kind) -> (con -> con)
               -> kind -> kind  *)
    fun map_kind fconvarkind fcon knd =
        let val fkind = map_kind fconvarkind fcon
        in
            case knd of
                KndType => knd
              | KndPi (x, k1, k2) =>
                    let val k1 = fkind k1
                        val (x, k2) = fconvarkind (x, k2)
                    in
                        KndPi (x, k1, k2)
                    end
              | KndUnit => knd
              | KndSigma (x, k1, k2) =>
                    let val k1 = fkind k1
                        val (x, k2) = fconvarkind (x, k2)
                    in
                        KndSigma (x, k1, k2)
                    end
              | KndSing c => KndSing (fcon c)
        end

    (* map_con : (convar -> con)
              -> (convar * con -> convar * con)
              -> (kind -> kind)
              -> con -> con *)
    (* XXX put these args in a record?  and for other map_* functions ... *)
    fun map_con fconvar fconvarcon fkind con =
            (* XXX jooc, better to do "fun fcon con = ... con"
                   or  "val fcon = ..." ?  optimization of partial appl.? *)
        let fun fcon con = map_con fconvar fconvarcon fkind con
            fun fconvarkindcon (x, k, c) =
                    let val k = fkind k
                        val (x, c) = fconvarcon (x, c)
                    in
                        (x, k, c)
                    end
        in
            case con of
                ConVar x => fconvar x
              | ConAbs xkc => ConAbs (fconvarkindcon xkc)
              | ConApp (c1, c2) => ConApp (fcon c1, fcon c2)
              | ConUnit => ConUnit
              | ConPair (c1, c2) => ConPair (fcon c1, fcon c2)
              | ConProj1 c => ConProj1 (fcon c)
              | ConProj2 c => ConProj2 (fcon c)
              | ConMu xkc => ConMu (fconvarkindcon xkc)
              | TyLit lt => con
              | TyArrow (c1, c2) => TyArrow (fcon c1, fcon c2)
              | TyRef c => TyRef (fcon c)
              | TyTagged => con
              | TyTag c => TyTag (fcon c)
              | TySum cons => TySum (map fcon cons)
              | TyProd cons => TyProd (map fcon cons)
              | TyForall xkc => TyForall (fconvarkindcon xkc)
        end

    (* map_exp : (expvar -> exp)
              -> (fbnd list -> fbnd list)
              -> (expvar * exp -> expvar * exp)
              -> (convar * exp -> convar * exp)
              -> (con -> con)
              -> (kind -> kind)
              -> exp -> exp *)
    fun map_exp fexpvar
                ffbnds
                fexpvarexp
                fconvarexp
                fcon
                fkind
                exp =
        let val fexp = map_exp fexpvar ffbnds fexpvarexp fconvarexp fcon fkind
            fun fconvarkindexp (x, k, e) =
                let val k = fkind k
                    val (x, e) = fconvarexp (x, e)
                in
                    (x, k, e)
                end
            fun fexpvarexpexp (x, e1, e2) =
                let val e1 = fexp e1
                    val (x, e2) = fexpvarexp (x, e2)
                in
                    (x, e1, e2)
                end
            fun fexpvarconexp (x, c, e) =
                let val c = fcon c
                    val (x, e) = fexpvarexp (x, e)
                in
                    (x, c, e)
                end
        in
            case exp of
                ExLit _ => exp
              | ExLam(x, c, e) => ExLam (fexpvarconexp (x, c, e))
              | ExUnOp(oper, e) => ExUnOp(oper, fexp e)
              | ExBinOp(e1, oper, e2) => ExBinOp(fexp e1, oper, fexp e2)
              | ExVar x => fexpvar x
              | ExFix fbnds => ExFix (ffbnds fbnds)
              | ExApp (e1, e2) => ExApp (fexp e1, fexp e2)
              | ExProd exps => ExProd (map fexp exps)
              | ExProj (i, e) => ExProj (i, fexp e)
              | ExHandle (e1, e2) => ExHandle (fexp e1, fexp e2)
              | ExRaise (c, e) => ExRaise (fcon c, fexp e)
              | ExRef e => ExRef (fexp e)
              | ExGet e => ExGet (fexp e)
              | ExSet (e1, e2) => ExSet (fexp e1, fexp e2)
              | ExRoll (c, e) => ExRoll (fcon c, fexp e)
              | ExUnroll e => ExUnroll (fexp e)
              | ExSumInj (c, i, e) => ExSumInj (fcon c, i, fexp e)
              | ExCase (c, e, es) => ExCase (fcon c, fexp e, map fexp es)
              | ExTagInj (e1, e2) => ExTagInj (fexp e1, fexp e2)
              | ExNewTag c => ExNewTag (fcon c)
              | ExIfTagOf (e1, e2, e3, e4) => ExIfTagOf (fexp e1, fexp e2,
                                                         fexp e3, fexp e4)
              | ExConAbs xke => ExConAbs (fconvarkindexp xke)
              | ExConApp (e, c) => ExConApp (fexp e, fcon c)
              | ExLet xe1e2 => ExLet (fexpvarexpexp xe1e2)
        end


    (*** replacement of convars ***)

    fun replace_con_kind convar convar' knd =
        map_kind (fn (cx, knd) =>
                    if Variable.eq (cx, convar) then
                        (* no free occurrences of var in knd -- bound by cx *)
                        (cx, knd)
                    else
                        (* cx not eq convar, so convar might be free in knd *)
                        (cx, replace_con_kind convar convar' knd))
                 (fn con => replace_con_con convar convar' con)
                 knd

    and replace_con_con convar convar' con =
        map_con (fn cx =>
                    if Variable.eq (cx, convar) then
                        ConVar convar'
                    else
                        ConVar cx)
                (fn (cx, c) =>
                    if Variable.eq (cx, convar) then
                        (cx, c)
                    else
                        (cx, replace_con_con convar convar' c))
                (replace_con_kind convar convar')
                con

    and replace_con_exp convar convar' exp =
        map_exp (fn x => ExVar x)
                (fn fbnds =>
                    map (fn (f, x, c1, c2, e) =>
                            (f, x, replace_con_con convar convar' c1,
                                   replace_con_con convar convar' c2,
                                   replace_con_exp convar convar' e))
                        fbnds)
                (fn (ex, e) => (ex, replace_con_exp convar convar' e))
                (fn (cx, e) =>
                    if Variable.eq (cx, convar) then
                        (cx, e)
                    else
                        (cx, replace_con_exp convar convar' e))
                (replace_con_con convar convar')
                (replace_con_kind convar convar')
                exp

    fun replace_exp_exp expvar expvar' exp =
        map_exp (fn x =>
                    if Variable.eq (x, expvar) then
                        ExVar expvar'
                    else
                        ExVar x)
                (fn fbnds =>
                    if List.exists
                        (fn (f, _, _, _, _) => Variable.eq (f, expvar))
                        fbnds
                    then
                        (* expvar is bound in entire phrase, do nothing *)
                        fbnds
                    else
                        map (fn (f, x, c1, c2, e) =>
                                let val e =
                                    if Variable.eq (x, expvar) then
                                        (* expvar is bound in e, do nothing *)
                                        e
                                    else
                                        replace_exp_exp expvar expvar' e
                                in

                                    (f, x, c1, c2, e)
                                end)
                            fbnds)
                (fn (ex, e) =>
                    if Variable.eq (ex, expvar) then
                        (ex, e)
                    else
                        (ex, replace_exp_exp expvar expvar' e))
                (fn (cx, e) => (cx, replace_exp_exp expvar expvar' e))
                (fn c => c) (* expvars do not appear in cons *)
                (fn k => k) (* expvars do not appear in kinds *)
                exp



    fun alpha_vary (c1var, c2obj) replace =
        let val c1var' = Variable.freshen c1var
            val c2obj' = replace c1var c1var' c2obj
        in
            (c1var', c2obj')
        end

    (* alpha_vary2 : ('a var * 'b) -> ('a var * 'b)
                  -> ('a var -> 'a var -> 'b -> 'b)
                  -> ('a var * ('b * 'b))

       varies two binding constructs to use the same (fresh) variable.
       for computing alpha equality *)
    fun alpha_vary2 (varA, objA) (varB, objB) replace =
        let val var' = Variable.freshen varA
            val objA' = replace varA var' objA
            val objB' = replace varB var' objB
        in
            (var', (objA', objB'))
        end


    (*** substitution for convars ***)

    fun subst_con_kind con convar knd =
        map_kind (fn (cx, k) =>
                    if Variable.eq (cx, convar) then
                        (cx, k)
                    else
                        let (* eagerly alpha vary with a fresh variable *)
                            val (cx, k) = alpha_vary (cx, k) replace_con_kind
                            (* cx \not\in fv(con), so we won't incur capture *)
                            val k = subst_con_kind con convar k
                        in
                            (cx, k)
                        end)
                 (subst_con_con con convar)
                 knd

    and subst_con_con confor convar incon =
        map_con (fn cx =>
                    if Variable.eq (cx, convar) then
                        (* cx is convar free, substitute for it *)
                        confor
                    else
                        (* not convar -- put it back *)
                        ConVar cx)
                (fn (cx, c) =>
                    if Variable.eq (cx, convar) then
                        (* cx is convar being rebound, leave it alone *)
                        (cx, c)
                    else
                        (* different variable, eagerly rename and continue *)
                        let val (cx, c) = alpha_vary (cx, c) replace_con_con
                            (* now, cx \not\in fv(confor) *)
                            val c = subst_con_con confor convar c
                        in
                            (cx, c)
                        end)
                (subst_con_kind confor convar)
                incon

    and subst_con_exp con convar exp =
        map_exp (fn x => ExVar x)
                (fn fbnds =>
                    map (fn (f, x, c1, c2, e) =>
                            (f, x, subst_con_con con convar c1,
                                   subst_con_con con convar c2,
                                   subst_con_exp con convar e))
                        fbnds)
                (fn (ex, e) => (ex, subst_con_exp con convar e))
                (fn (cx, e) =>
                    if Variable.eq (cx, convar) then
                        (cx, e)
                    else
                        let val (cx, e) = alpha_vary (cx, e) replace_con_exp
                            val e = subst_con_exp con convar e
                        in
                            (cx, e)
                        end)
                (subst_con_con con convar)
                (subst_con_kind con convar)
                exp

    fun subst_con_cons confor convar consin =
        map (subst_con_con confor convar) consin


    fun labelledKndSing con knd =
        case knd of
            KndType => KndSing con
          | KndSing _ => KndSing con
          | KndPi (cx, k1, k2) =>
                KndPi (cx, k1, labelledKndSing (ConApp (con, ConVar cx)) k2)
          | KndUnit => KndUnit
          | KndSigma (cx, k1, k2) =>
                let val k2' = subst_con_kind (ConProj1 con) cx k2
                in
                    kndPair (labelledKndSing (ConProj1 con) k1,
                             labelledKndSing (ConProj2 con) k2')
                end
  end
